//+------------------------------------------------------------------+
//|                                StepWiseFeatureSelection_Demo.mq5 |
//|                                  Copyright 2024, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2024, MetaQuotes Ltd."
#property link      "https://www.mql5.com"
#property version   "1.00"
#property script_show_inputs
#include<OLS.mqh>
#include<stepwise.mqh>

input ulong NumKeptVars = 1;
input ulong NumFolds = 10;
input ulong MinNumVars = 3;
input ulong MaxNumVars = 5;
input long NumReplications = 1;
//+------------------------------------------------------------------+
//|    ols base feature selection                                    |
//+------------------------------------------------------------------+
class CStepwise:public CStepwisebase
  {
private :
   OLS               *m_ols;
   //+------------------------------------------------------------------+
   //|    fit the OLS model                                             |
   //+------------------------------------------------------------------+
   virtual bool      fitModel(ulong num_preds,ulong row_start,ulong row_stop,vector& targets)
     {
      int k1,k2;

      ulong nvars = num_preds + num_preds * (num_preds+1)/2;
      ulong ntrain = m_samples - (row_stop - row_start);

      vector dependent(ntrain);
      matrix independent(ntrain,nvars+1);
      ntrain = 0;
      
      for(ulong sample=0; sample<m_samples; sample++)
        {
         if(sample>=row_start && sample<row_stop)
            continue;
         k1=0;
         k2=1;
         for(ulong var=0; var<nvars; var++)
           {
            if(var<num_preds)
               independent[ntrain][var]=m_data[sample][m_trial_vars[var]];
            else
              {
               if(var<(num_preds*2))
                 {
                  independent[ntrain][var]=pow(m_data[sample][m_trial_vars[var-num_preds]],2.0);
                 }
               else
                 {
                  independent[ntrain][var]=m_data[sample][m_trial_vars[k1]]*m_data[sample][m_trial_vars[k2]];
                  ++k2;
                  if(ulong(k2)==num_preds)
                    {
                     ++k1;
                     k2 = k1 + 1;
                    }
                 }
              }
           }
         independent[ntrain][nvars] = 1.0;
         dependent[ntrain++] = targets[sample];
        }

      return m_ols.Fit(dependent,independent);
     }
   //+------------------------------------------------------------------+
   //|      evaluate the model                                          |
   //+------------------------------------------------------------------+
   virtual double    evalModel(ulong num_preds,ulong row_start,ulong row_stop,vector& targets)
     {
      if(m_ols.Loglikelihood()==EMPTY_VALUE)
        {
         Print(__FUNCTION__, " OLS error ");
         return EMPTY_VALUE;
        }

      int k1,k2;

      ulong nvars = num_preds + num_preds * (num_preds+1)/2;
      double prediction;
      vector row(nvars);
      double error = 0.0;
      for(ulong sample=row_start; sample<row_stop; sample++)
        {
         k1=0;
         k2=1;
         for(ulong var=0; var<nvars; var++)
           {
            if(var<num_preds)
               row[var]=m_data[sample][m_trial_vars[var]];
            else
              {
               if(var<(num_preds*2))
                 {
                  row[var]=pow(m_data[sample][m_trial_vars[var-num_preds]],2.0);
                 }
               else
                 {
                  row[var]=m_data[sample][m_trial_vars[k1]]*m_data[sample][m_trial_vars[k2]];
                  ++k2;
                  if(ulong(k2)==num_preds)
                    {
                     ++k1;
                     k2 = k1 + 1;
                    }
                 }
              }
           }
           
         prediction = m_ols.Predict(row);
         if(prediction == EMPTY_VALUE)
          {
            Print(__FUNCTION__, " OLS predict() error ");
            return EMPTY_VALUE;
          }
         error+=pow(prediction-targets[sample],2.0);
        }
       
       return error;
     }
public :
                     CStepwise(void)
     {
      m_ols = new OLS();
     }
                    ~CStepwise(void)
     {
      if(CheckPointer(m_ols) == POINTER_DYNAMIC)
          delete m_ols;
     }
  };
//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
void OnStart()
  {
//---
   MathSrand(120);
//---
   matrix mat(100,10);
//---
   mat.Random(0.0,1.0);
//---
   vector target = mat.Col(0) + mat.Col(5) + mat.Col(4) + mat.Col(6);
//---
   CStepwise stepwise;
   stepwise.setParams(NumKeptVars,NumFolds,MinNumVars,MaxNumVars,NumReplications,true);
   if(!stepwise.stepwiseSelection(mat,target))
      return;
/*      
   ulong subset[];
   if(!stepwise.getFeatureSetAt(3,subset))
     return;
     
   Print(" selected variables are :");
   ArrayPrint(subset);
*/   
  }
//+------------------------------------------------------------------+
